@@ -4,12 +4,23 @@ import inspect
 import re
 from itertools import chain
 
-from six import string_types
-
 from .actions import Action
 from .event import Event
+from .util import StringType
 
-ALLOWED_KEYS = tuple(i for i in Event.__dict__.keys() if not i.startswith('_') and i not in ('tracer', 'thread'))
+__all__ = (
+    'And',
+    'From',
+    'Not',
+    'Or',
+    'Query',
+    'When',
+)
+
+ALLOWED_KEYS = tuple(sorted(
+    i for i in Event.__dict__.keys()
+    if not i.startswith('_') and i not in ('tracer', 'thread', 'frame')
+))
 ALLOWED_OPERATORS = (
     'startswith', 'endswith', 'in', 'contains', 'regex',
     'sw', 'ew', 'has', 'rx',
@@ -17,18 +28,11 @@ ALLOWED_OPERATORS = (
 )
 
 
-def _sloppy_hash(obj):
-    try:
-        return hash(obj)
-    except TypeError:
-        return 'id(%x)' % id(obj)
-
-
 class Query(object):
     """
     A query class.
 
-    See :class:`hunter.Event` for fields that can be filtered on.
+    See :class:`hunter.event.Event` for fields that can be filtered on.
     """
     def __init__(self, **query):
         """
@@ -75,13 +79,13 @@ class Query(object):
             elif count == 2:
                 prefix, operator = parts
                 if operator in ('startswith', 'sw'):
-                    if not isinstance(value, string_types):
+                    if not isinstance(value, StringType):
                         if not isinstance(value, (list, set, tuple)):
                             raise ValueError('Value %r for %r is invalid. Must be a string, list, tuple or set.' % (value, key))
                         value = tuple(value)
                     mapping = query_startswith
                 elif operator in ('endswith', 'ew'):
-                    if not isinstance(value, string_types):
+                    if not isinstance(value, StringType):
                         if not isinstance(value, (list, set, tuple)):
                             raise ValueError('Value %r for %r is invalid. Must be a string, list, tuple or set.' % (value, key))
                         value = tuple(value)
@@ -294,9 +298,79 @@ class When(object):
     __rand__ = __and__
 
 
+class From(object):
+    """
+    From-point filtering mechanism. Switches on to running the predicate after condition maches, and switches off when
+    the depth returns to the same level.
+
+    After ``condition(event)`` returns ``True`` the ``event.depth`` will be saved and calling this object with an
+    ``event`` will return ``predicate(event)`` until ``event.depth - watermark`` is equal to the depth that was saved.
+
+    Args:
+        condition (callable): A callable that returns True/False or a :class:`hunter.predicates.Query` object.
+        predicate (callable): Optional callable that returns True/False or a :class:`hunter.predicates.Query` object to
+            run after ``condition`` first returns ``True``.
+        watermark (int): Depth difference to switch off and wait again on ``condition``.
+    """
+
+    def __init__(self, condition, predicate=None, watermark=0):
+        self.condition = condition
+        self.predicate = predicate
+        self.watermark = watermark
+        self.waiting_for_condition = True
+        self.depth = -1
+
+    def __str__(self):
+        return 'From(%s, %s)' % (self.condition, self.predicate)
+
+    def __repr__(self):
+        return '<hunter.predicates.From: condition=%r, predicate=%r, watermark=%r>' % (
+            self.condition, self.predicate, self.watermark
+        )
+
+    def __eq__(self, other):
+        return (
+            isinstance(other, From)
+            and self.condition == other.condition
+            and self.predicate == other.predicate
+        )
+
+    def __call__(self, event):
+        """
+        Handles the event.
+        """
+        if event.depth - self.watermark <= self.depth:
+            self.waiting_for_condition = True
+            self.depth = -1
+
+        if self.waiting_for_condition:
+            if self.condition(event):
+                self.waiting_for_condition = False
+                self.depth = event.depth
+            else:
+                return False
+
+        if self.predicate is None:
+            return True
+        else:
+            return self.predicate(event)
+
+    def __or__(self, other):
+        return From(Or(self.condition, other), self.predicate)
+
+    def __and__(self, other):
+        return From(self.condition, And(self.predicate, other))
+
+    def __invert__(self):
+        return Not(self)
+
+    __ror__ = __or__
+    __rand__ = __and__
+
+
 class And(object):
     """
-    `And` predicate. Exits at the first sub-predicate that returns ``False``.
+    Returns ``False`` at the first sub-predicate that returns ``False``, otherwise returns ``True``.
     """
 
     def __init__(self, *predicates):
@@ -342,7 +416,7 @@ class And(object):
 
 class Or(object):
     """
-    `Or` predicate. Exits at first sub-predicate that returns ``True``.
+    Returns ``True`` after the first sub-predicate that returns ``True``.
     """
 
     def __init__(self, *predicates):
@@ -388,7 +462,7 @@ class Or(object):
 
 class Not(object):
     """
-    `Not` predicate.
+    Simply returns ``not predicate(event)``.
     """
     def __init__(self, predicate):
         self.predicate = predicate
